//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Import/SpecularDataImportWidget.cpp
//! @brief     Implements class SpecularDataImportWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Import/SpecularDataImportWidget.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Support/IO/IDataLoaderResultModel.h"
#include "GUI/View/Loaders/DataLoaderUtil.h"
#include "GUI/View/Loaders/QREDataLoader.h"
#include "GUI/View/PlotSpecular/SpecularDataCanvas.h"
#include "GUI/View/Widget/StyledToolbar.h"
#include <QAction>

SpecularDataImportWidget::SpecularDataImportWidget(QWidget* parent)
    : DataAccessWidget(parent)
    , m_loader(nullptr)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setGeometry(0, 0, 861, 580);

    auto* vbox = new QVBoxLayout;
    setLayout(vbox);

    auto* splitter = new QSplitter(Qt::Horizontal);
    vbox->addWidget(splitter);
    splitter->setChildrenCollapsible(false);

    //... Left side of splitter: import filter settings

    auto* importer_widget = new QWidget;
    splitter->addWidget(importer_widget);

    auto* importer_vbox = new QVBoxLayout;
    importer_widget->setLayout(importer_vbox);
    importer_vbox->setSpacing(20);

    auto* import_settings_widget = new QWidget;
    importer_vbox->addWidget(import_settings_widget);
    import_settings_widget->setWindowTitle("Import settings");

    auto* import_settings_vbox = new QVBoxLayout;
    import_settings_widget->setLayout(import_settings_vbox);

    m_properties_widget = new QWidget; // to be populated later
    import_settings_vbox->addWidget(m_properties_widget);

    //... Right side of splitter: data view

    auto* tab = new QTabWidget;
    splitter->addWidget(tab);

    auto* data_widget = new QWidget;
    tab->addTab(data_widget, "Table");

    auto* data_vbox = new QVBoxLayout;
    data_widget->setLayout(data_vbox);

    m_data_table = new QTableView;
    data_vbox->addWidget(m_data_table);

    auto* graph_widget = new QWidget;
    tab->addTab(graph_widget, "Graph");

    auto* graph_vbox = new QVBoxLayout;
    graph_widget->setLayout(graph_vbox);

    auto* plotToolbar = new StyledToolbar;
    graph_vbox->addWidget(plotToolbar);

    m_canvas = new SpecularDataCanvas;
    graph_vbox->addWidget(m_canvas);

    //...

    updatePropertiesEdits();

    connect(m_canvas->customPlot(), &QCustomPlot::axisClick, this,
            &SpecularDataImportWidget::onPlotAxisClicked);

    m_canvas->enableDeprecatedOnMousePress(false); // we have an own handler

    plotToolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    for (auto* action : m_canvas->actionList())
        plotToolbar->addAction(action);
}

void SpecularDataImportWidget::setJobOrRealItem(QObject* real_item)
{
    DataAccessWidget::setJobOrRealItem(real_item);
    ASSERT(realItem());
    m_canvas->setJobOrRealItem(real_item);

    m_loader = dynamic_cast<IDataLoader*>(realItem()->dataLoader());
    ASSERT(m_loader); // only items which have a loader are allowed for this widget. Every other
                      // items do not support this widget

    updatePropertiesEdits();
    updatePreview();
    connect(m_loader, &IDataLoader::importSettingsChanged, this,
            &SpecularDataImportWidget::onPropertiesChanged, Qt::UniqueConnection);
}

QList<QAction*> SpecularDataImportWidget::actionList()
{
    return {};
}

void SpecularDataImportWidget::onContextMenuRequest(const QPoint& point)
{
    QMenu menu;
    for (auto* action : actionList())
        menu.addAction(action);
    menu.exec(point);
}

void SpecularDataImportWidget::onPlotAxisClicked(QCPAxis* axis, QCPAxis::SelectablePart /*part*/,
                                                 QMouseEvent* event)
{
    if (event->button() == Qt::RightButton && axis->axisType() == QCPAxis::atLeft) {
        QMenu menu;

        auto* lin = new QAction("Linear");
        connect(lin, &QAction::triggered, [this]() { realSpecularDataItem()->setLog(false); });
        lin->setCheckable(true);
        lin->setChecked(!realSpecularDataItem()->isLog());

        auto* log = new QAction("Logarithmic");
        connect(log, &QAction::triggered, [this]() { realSpecularDataItem()->setLog(true); });
        log->setCheckable(true);
        log->setChecked(realSpecularDataItem()->isLog());

        auto* ag = new QActionGroup(&menu);
        ag->addAction(lin);
        ag->addAction(log);

        menu.addAction(lin);
        menu.addAction(log);

        menu.exec(event->globalPos());
    }
}

void SpecularDataImportWidget::updatePropertiesEdits()
{
    for (auto* child : m_properties_widget->children())
        delete child;

    if (m_properties_widget->layout())
        delete m_properties_widget->layout();

    if (m_loader)
        m_loader->populateImportSettingsWidget(m_properties_widget);

    const bool hasChildren = !m_properties_widget->children().empty();

    m_properties_widget->setVisible(hasChildren);
}

void SpecularDataImportWidget::updatePreview()
{
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (m_loader) {
        auto* oldModel = m_data_table->selectionModel(); // sic!! according to Qt docu
                                                         // of QAbstractItemView::setModel
        auto* resultModel = m_loader->createResultModel();
        if (resultModel != nullptr) {
            const auto processedSections =
                resultModel->sectionsOfColumnType(IDataLoaderResultModel::ColumnType::processed);

            m_data_table->setModel(resultModel);

            // if the result model has a line column, then do not show the vertical header view
            const bool hasLinesColumn =
                !resultModel->sectionsOfColumnType(IDataLoaderResultModel::ColumnType::line)
                     .isEmpty();
            m_data_table->verticalHeader()->setHidden(hasLinesColumn);

            m_data_table->resizeColumnsToContents();
        } else
            m_data_table->setModel(nullptr);

        delete oldModel;
    }

    QApplication::restoreOverrideCursor();
}

void SpecularDataImportWidget::onPropertiesChanged()
{
    m_loader->applyImportSettings(); // #baimport: may be duplicate
    QApplication::setOverrideCursor(Qt::WaitCursor);
    m_loader->processContents();
    QApplication::restoreOverrideCursor();

    // If there is a linked instrument, any change in import settings can break the compatibility.
    // Therefore check the compatibility and break the link  if necessary
    ASSERT(gProjectDocument.has_value());

    const auto* linkedInstrument =
        gProjectDocument.value()->instrumentModel()->findInstrumentItemById(
            realItem()->instrumentId());

    if (linkedInstrument)
        if (!linkedInstrument->alignedWith(realItem()))
            realItem()->unlinkFromInstrument();

    updatePreview();
}

QString SpecularDataImportWidget::currentFileName() const
{
    return realItem()->nativeFileName();
}
